@@ -17,3 +17,6 @@
 #undef HAVE_FRIBIDI
 #undef HAVE_UNISTD_H
 #undef HAVE_FREETYPE_2_0_3
+
+#undef MINI_XFTCONFIG_DIR
+#undef HAVE_XFT2
