@@ -28,11 +28,7 @@ struct _ThaiFontInfo
 {
   PangoFont       *font;
   ThaiFontSet      font_set;
-  ThaiFontInfoType info_type;
-  union {
-    PangoXSubfont subfont; /* THAI_FONTINFO_X */
-    gpointer      face;    /* THAI_FONTINFO_XFT */
-  } info;
+  PangoXSubfont    subfont; /* For X backend */
 };
 
 /*
