@@ -49,6 +49,46 @@ typedef struct _PangoXContextInfo PangoXContextInfo;
 typedef struct _PangoXLigatureInfo   PangoXLigatureInfo;
 typedef struct _PangoXLigatureSource PangoXLigatureSource;
 
+#ifndef HAVE_STRTOK_R
+/* This implementation of strtok_r comes from the GNU C library.
+ * Copyright (C) 1991, 1996, 1997, 1998, 1999, 2001 Free Software Foundation, Inc.
+ */
+static char *
+my_strtok_r (char       *s,
+	     const char *delim,
+	     char      **save_ptr)
+{
+  char *token;
+
+  if (s == NULL)
+    s = *save_ptr;
+
+  /* Scan leading delimiters.  */
+  s += strspn (s, delim);
+  if (*s == '\0')
+    {
+      *save_ptr = s;
+      return NULL;
+    }
+
+  /* Find the end of the token.  */
+  token = s;
+  s = strpbrk (token, delim);
+  if (s == NULL)
+    /* This token finishes the string.  */
+    *save_ptr = token + strlen (token);
+  else
+    {
+      /* Terminate the token and make *SAVE_PTR point past it.  */
+      *s = '\0';
+      *save_ptr = s + 1;
+    }
+  return token;
+}
+#else
+#define my_strtok_r strtok_r
+#endif /* HAVE_STRTOK_R */
+
 static int
 hex_to_integer (const char *s) 
 {
@@ -73,7 +113,7 @@ parse_gintset_spec (char *s)
 {
   char *m = NULL;
   PangoIntSet *set = pango_int_set_new ();
-  s = strtok_r (s, ",", &m);
+  s = my_strtok_r (s, ",", &m);
   while (s)
     {
       char *p = strchr (s, '-');
@@ -93,7 +133,7 @@ parse_gintset_spec (char *s)
 	  if (start != -1 && end != -1)
 	    pango_int_set_add_range (set, start, end);
 	}
-      s = strtok_r (NULL, ",", &m);
+      s = my_strtok_r (NULL, ",", &m);
     }
   return set;
 }
@@ -1046,7 +1086,8 @@ pango_x_insert_subfont (PangoFont *font, const char *xlfd)
  * @subfont_charsets: location to store a pointer to an array of subfont IDs for each found subfont;
  *               the result must be freed using g_free().
  * 
- * Lists the subfonts of a given font.
+ * Lists the subfonts of a given font. The result is ordered first by charset,
+ * and then within each charset, by the order of fonts in the font specification.
  *
  * Return value: length of the arrays stored in @subfont_ids and 
  * @subfont_charsets.
@@ -1124,8 +1165,8 @@ pango_x_list_subfonts (PangoFont        *font,
 
   n_subfonts = 0;
 
-  for (i=0; i<xfont->n_fonts; i++)
-    for (j=0; j<n_charsets; j++)
+  for (j=0; j<n_charsets; j++)
+    for (i=0; i<xfont->n_fonts; i++)
       if (subfont_lists[j][i])
 	{
 	  (*subfont_ids)[n_subfonts] = subfont_lists[j][i];
@@ -1724,7 +1765,7 @@ font_struct_get_ligatures (PangoFontMap *fontmap,
             {
               char *val = g_strdup (pango_x_fontmap_name_from_atom (fontmap, fs->properties[i].card32));
               char *p;
-              char *a = strtok_r (val, " ", &p);
+              char *a = my_strtok_r (val, " ", &p);
               while (a)
                 {
                   char *r;
@@ -1805,7 +1846,7 @@ font_struct_get_ligatures (PangoFontMap *fontmap,
                         *r = 0;
                         r++;
                         q = a;
-                        q = strtok_r (q, "+", &m);
+                        q = my_strtok_r (q, "+", &m);
                         while (q) 
                           {
                             n_source ++;
@@ -1839,10 +1880,10 @@ font_struct_get_ligatures (PangoFontMap *fontmap,
                                 source [n_source-1].is_set = 0;
                                 source [n_source-1].data.glyph = i;
                               }
-                            q = strtok_r (NULL, "+", &m);
+                            q = my_strtok_r (NULL, "+", &m);
                           }
                         q = r;
-                        q = strtok_r (q, "+", &m);
+                        q = my_strtok_r (q, "+", &m);
                         while (q) 
                           {
                             n_dest++;
@@ -1873,7 +1914,7 @@ font_struct_get_ligatures (PangoFontMap *fontmap,
 				  }
 			      }
                             
-                            q = strtok_r (NULL, "+", &m);
+                            q = my_strtok_r (NULL, "+", &m);
                           }
                         
                         xli = linfo + n_linfo - 1;
@@ -1892,7 +1933,7 @@ font_struct_get_ligatures (PangoFontMap *fontmap,
                     }
                   
                   /* end switch */
-                  a = strtok_r (NULL, " ", &p);
+                  a = my_strtok_r (NULL, " ", &p);
                 }
               g_free (val);
             }
