@@ -24,8 +24,6 @@
 #include <stdlib.h>
 
 #include "pangoxft-private.h"
-#include "X11/Xft/XftFreetype.h"
-
 #include "pango-layout.h"
 #include "pango-modules.h"
 #include "pango-utils.h"
@@ -71,6 +69,7 @@ static void                  pango_xft_font_get_glyph_extents (PangoFont
 static PangoFontMetrics *    pango_xft_font_get_metrics       (PangoFont        *font,
 							       PangoLanguage    *language);
 
+static XftFont *xft_font_get_font (PangoFont *font);
 
 GType
 pango_xft_font_get_type (void)
@@ -127,7 +126,7 @@ pango_xft_font_class_init (PangoXftFontClass *class)
 
 PangoXftFont *
 _pango_xft_font_new (PangoFontMap               *fontmap,
-		     XftPattern                 *pattern)
+		     FcPattern			*pattern)
 {
   PangoXftFont *xfont;
 
@@ -153,6 +152,8 @@ get_mini_font (PangoFont *font)
 {
   PangoXftFont *xfont = (PangoXftFont *)font;
 
+  g_assert (xfont->fontmap);
+
   if (!xfont->mini_font)
     {
       Display *display;
@@ -161,7 +162,6 @@ get_mini_font (PangoFont *font)
       int width = 0, height = 0;
       XGlyphInfo extents;
       XftFont *mini_xft;
-      FT_Face face;
       
       _pango_xft_font_map_get_info (xfont->fontmap, &display, NULL);
 
@@ -172,19 +172,16 @@ get_mini_font (PangoFont *font)
       xfont->mini_font = pango_font_map_load_font (xfont->fontmap, NULL, desc);
       pango_font_description_free (desc);
       
-      mini_xft = pango_xft_font_get_font (xfont->mini_font);
-      face = pango_xft_font_get_face (xfont->mini_font);
+      mini_xft = xft_font_get_font (xfont->mini_font);
       
       for (i = 0 ; i < 16 ; i++)
 	{
 	  char c = i < 10 ? '0' + i : 'A' + i - 10;
-	  XftChar32 glyph = FT_Get_Char_Index (face, c);
-	  
-	  XftTextExtents32 (display, mini_xft, &glyph, 1, &extents);
-
+	  XftTextExtents8 (display, mini_xft, (FcChar8 *) &c, 1, &extents);
 	  width = MAX (width, extents.width);
 	  height = MAX (height, extents.height);
 	}
+
       
       xfont->mini_width = width;
       xfont->mini_height = height;
@@ -261,9 +258,28 @@ pango_xft_real_render (Display          *display,
 		       gint              y)
 {
   PangoXftFont *xfont = PANGO_XFT_FONT (font);
-  XftFont *xft_font = pango_xft_font_get_font (font);
+  XftFont *xft_font = xft_font_get_font (font);
   int i;
   int x_off = 0;
+#define N_XFT_LOCAL	1024
+  XftGlyphSpec  xft_glyphs[N_XFT_LOCAL];
+  XftCharSpec	chars[4];     /* for unknown */
+  int		n_xft_glyph = 0;
+
+  if (!xfont->fontmap)		/* Display closed */
+    return;
+
+#define FLUSH_GLYPHS() G_STMT_START {						\
+  if (n_xft_glyph)								\
+    {										\
+       if (draw)								\
+         XftDrawGlyphSpec (draw, color, xft_font, xft_glyphs, n_xft_glyph);	\
+       else									\
+         XftGlyphSpecRender (display, PictOpOver, src_picture, xft_font,	\
+	  		     dest_picture, 0, 0, xft_glyphs, n_xft_glyph);	\
+       n_xft_glyph = 0;								\
+    }										\
+  } G_STMT_END
 
   /* Slow initial implementation. For speed, it should really
    * collect the characters into runs, and draw multiple
@@ -286,11 +302,10 @@ pango_xft_real_render (Display          *display,
 	      int j, k;
 	      
 	      PangoFont *mini_font = get_mini_font (font);
-	      XftFont *mini_xft = pango_xft_font_get_font (mini_font);
-	      FT_Face face = pango_xft_font_get_face (xfont->mini_font);
+	      XftFont *mini_xft = xft_font_get_font (mini_font);
       
 	      glyph &= ~PANGO_XFT_UNKNOWN_FLAG;
-
+	      
 	      ys[0] = y + PANGO_PIXELS (glyphs->glyphs[i].geometry.y_offset) - xft_font->ascent + (xft_font->ascent + xft_font->descent - xfont->mini_height * 2 - xfont->mini_pad * 5) / 2;
 	      ys[1] = ys[0] + 2 * xfont->mini_pad + xfont->mini_height;
 	      ys[2] = ys[1] + xfont->mini_height + xfont->mini_pad;
@@ -306,38 +321,42 @@ pango_xft_real_render (Display          *display,
 	      
 	      g_snprintf (buf, sizeof(buf), "%04X", glyph);
 
+	      FLUSH_GLYPHS ();
 	      for (j = 0; j < 2; j++)
 		for (k = 0; k < 2; k++)
 		  {
-		    XftChar32 glyph = FT_Get_Char_Index (face, buf[2*j + k]);
-		    if (draw)
-		      XftDrawString32 (draw, color, mini_xft,
-				       xs[k+1], ys[j+1], &glyph, 1);
-		    else
-		      XftRenderString32 (display, src_picture, mini_xft->u.ft.font, dest_picture, 0, 0,
-					 xs[k+1], ys[j+1], &glyph, 1);
-					 
+		    XftCharSpec *c = &chars[j * 2 + k];
+		    c->ucs4 = buf[j * 2 + k] & 0xff;
+		    c->x = xs[k+1];
+		    c->y = ys[j+1];
 		  }
+	      if (draw)
+		XftDrawCharSpec (draw, color, mini_xft,
+				 chars, 4);
+	      else
+		XftCharSpecRender (display, PictOpOver, src_picture,
+				   mini_xft, dest_picture, 0, 0,
+				   chars, 4);
 	    }
 	  else if (glyph)
 	    {
-	      if (draw)
-		XftDrawString32 (draw, color, xft_font,
-				 x + PANGO_PIXELS (x_off + glyphs->glyphs[i].geometry.x_offset),
-				 y + PANGO_PIXELS (glyphs->glyphs[i].geometry.y_offset),
-				 &glyph, 1);
-	      else
-		XftRenderString32 (display, src_picture, xft_font->u.ft.font, dest_picture, 0, 0,
-				   x + PANGO_PIXELS (x_off + glyphs->glyphs[i].geometry.x_offset),
-				   y + PANGO_PIXELS (glyphs->glyphs[i].geometry.y_offset),
-				   &glyph, 1);
+	      if (n_xft_glyph == N_XFT_LOCAL)
+		FLUSH_GLYPHS ();
+	      
+	      xft_glyphs[n_xft_glyph].x = x + PANGO_PIXELS (x_off + glyphs->glyphs[i].geometry.x_offset);
+	      xft_glyphs[n_xft_glyph].y = y + PANGO_PIXELS (glyphs->glyphs[i].geometry.y_offset);
+	      xft_glyphs[n_xft_glyph].glyph = glyph;
+	      n_xft_glyph++;
 	    }
 	}
       
       x_off += glyphs->glyphs[i].geometry.width;
     }
-}
+  
+  FLUSH_GLYPHS ();
 
+#undef FLUSH_GLYPHS
+}
 
 /**
  * pango_xft_render:
@@ -422,42 +441,48 @@ pango_xft_font_get_metrics (PangoFont     *font,
       PangoLayout *layout;
       PangoRectangle extents;
       PangoContext *context;
-      XftFont *xft_font = pango_xft_font_get_font (font);
+      XftFont *xft_font;
       Display *display;
-  
-      _pango_xft_font_map_get_info (xfont->fontmap, &display, NULL);
-      context = pango_xft_get_context (display, 0);
-
-      info = g_new (PangoXftMetricsInfo, 1);
-      info->sample_str = sample_str;
-      info->metrics = pango_font_metrics_new ();
-
-      info->metrics->ascent = PANGO_SCALE * xft_font->ascent;
-      info->metrics->descent = PANGO_SCALE * xft_font->descent;
-      info->metrics->approximate_char_width = 
-        info->metrics->approximate_digit_width = 
-        PANGO_SCALE * xft_font->max_advance_width;
 
+      info = g_new0 (PangoXftMetricsInfo, 1);
       xfont->metrics_by_lang = g_slist_prepend (xfont->metrics_by_lang, 
                                                 info);
 
-      pango_context_set_language (context, language);
-      layout = pango_layout_new (context);
-      pango_layout_set_font_description (layout, xfont->description);
-
-      pango_layout_set_text (layout, sample_str, -1);      
-      pango_layout_get_extents (layout, NULL, &extents);
+	  
+      if (xfont->fontmap)
+	{
+	  xft_font = xft_font_get_font (font);
       
-      info->metrics->approximate_char_width = 
-        extents.width / g_utf8_strlen (sample_str, -1);
+	  _pango_xft_font_map_get_info (xfont->fontmap, &display, NULL);
+	  context = pango_xft_get_context (display, 0);
 
-      pango_layout_set_text (layout, "0123456789", -1);      
-      pango_layout_get_extents (layout, NULL, &extents);
+	  info->sample_str = sample_str;
+	  info->metrics = pango_font_metrics_new ();
+
+	  info->metrics->ascent = PANGO_SCALE * xft_font->ascent;
+	  info->metrics->descent = PANGO_SCALE * xft_font->descent;
+	  info->metrics->approximate_char_width = 
+	    info->metrics->approximate_digit_width = 
+	    PANGO_SCALE * xft_font->max_advance_width;
+
+	  pango_context_set_language (context, language);
+	  layout = pango_layout_new (context);
+	  pango_layout_set_font_description (layout, xfont->description);
+
+	  pango_layout_set_text (layout, sample_str, -1);      
+	  pango_layout_get_extents (layout, NULL, &extents);
       
-      info->metrics->approximate_digit_width = extents.width / 10;
+	  info->metrics->approximate_char_width = 
+	    extents.width / g_utf8_strlen (sample_str, -1);
 
-      g_object_unref (G_OBJECT (layout));
-      g_object_unref (G_OBJECT (context));
+	  pango_layout_set_text (layout, "0123456789", -1);      
+	  pango_layout_get_extents (layout, NULL, &extents);
+      
+	  info->metrics->approximate_digit_width = extents.width / 10;
+
+	  g_object_unref (G_OBJECT (layout));
+	  g_object_unref (G_OBJECT (context));
+	}
     }
 
   return pango_font_metrics_ref (info->metrics);
@@ -489,27 +514,27 @@ static void
 pango_xft_font_finalize (GObject *object)
 {
   PangoXftFont *xfont = (PangoXftFont *)object;
-  Display *display;
-  
-  _pango_xft_font_map_get_info (xfont->fontmap, &display, NULL);
 
-  _pango_xft_font_map_remove (xfont->fontmap, xfont);
+  if (xfont->fontmap)
+    _pango_xft_font_map_remove (xfont->fontmap, xfont);
 
   if (xfont->mini_font)
     g_object_unref (xfont->mini_font);
 
-  if (xfont->ot_info)
-    g_object_unref (xfont->ot_info);
-
   pango_font_description_free (xfont->description);
 
   g_slist_foreach (xfont->metrics_by_lang, (GFunc)free_metrics_info, NULL);
   g_slist_free (xfont->metrics_by_lang);  
 
   if (xfont->xft_font)
-    XftFontClose (display, xfont->xft_font);
-  else
-    XftPatternDestroy (xfont->font_pattern); /* If we opened the font it will own and destroy this pattern */
+    {
+      Display *display;
+      
+      _pango_xft_font_map_get_info (xfont->fontmap, &display, NULL);
+      XftFontClose (display, xfont->xft_font);
+    }
+
+  FcPatternDestroy (xfont->font_pattern);
 
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
@@ -527,53 +552,8 @@ pango_xft_font_get_coverage (PangoFont     *font,
 			     PangoLanguage *language)
 {
   PangoXftFont *xfont = (PangoXftFont *)font;
-  char *filename = NULL;
-  FT_Face face;
-  PangoCoverage *coverage;
-  Display *display;
-  
-  _pango_xft_font_map_get_info (xfont->fontmap, &display, NULL);
 
-  XftPatternGetString (xfont->font_pattern, XFT_FILE, 0, &filename);
-  
-  coverage = _pango_xft_font_map_get_coverage (xfont->fontmap, filename);
-
-  if (coverage)
-    return pango_coverage_ref (coverage);
-
-  face = pango_xft_font_get_face (font);
-
-  coverage = pango_coverage_new ();
-#ifdef HAVE_FT_GET_FIRST_CHAR
-  {
-    FT_UInt gindex;
-    FT_ULong charcode;
-    
-    charcode = FT_Get_First_Char (face, &gindex);
-    while (gindex)
-      {
-	pango_coverage_set (coverage, charcode, PANGO_COVERAGE_EXACT);
-	charcode = FT_Get_Next_Char (face, charcode, &gindex);
-      }
-  }
-#else  
-  /* Ugh, this is going to be SLOW */
-  {
-    gunichar wc;
-    
-    for (wc = 0; wc < G_MAXUSHORT; wc++)
-      {
-	FT_UInt glyph = FT_Get_Char_Index (face, wc);
-	
-	if (glyph && glyph < face->num_glyphs)
-	  pango_coverage_set (coverage, wc, PANGO_COVERAGE_EXACT);
-      }
-  }
-#endif
-
-  _pango_xft_font_map_set_coverage (xfont->fontmap, filename, coverage);
- 
-  return coverage;
+  return _pango_xft_font_map_get_coverage (xfont->fontmap, xfont->font_pattern);
 }
 
 static void
@@ -583,10 +563,13 @@ pango_xft_font_get_glyph_extents (PangoFont        *font,
 				  PangoRectangle   *logical_rect)
 {
   PangoXftFont *xfont = (PangoXftFont *)font;
-  XftFont *xft_font = pango_xft_font_get_font (font);
+  XftFont *xft_font = xft_font_get_font (font);
   XGlyphInfo extents;
   Display *display;
 
+  if (!xfont->fontmap)		/* Display closed */
+    goto fallback;
+
   _pango_xft_font_map_get_info (xfont->fontmap, &display, NULL);
   
   if (glyph == (PangoGlyph)-1)
@@ -614,7 +597,8 @@ pango_xft_font_get_glyph_extents (PangoFont        *font,
     }
   else if (glyph)
     {
-      XftTextExtents32 (display, xft_font, &glyph, 1, &extents);
+      FT_UInt ft_glyph = glyph;
+      XftGlyphExtents (display, xft_font, &ft_glyph, 1, &extents);
 
       if (ink_rect)
 	{
@@ -634,6 +618,8 @@ pango_xft_font_get_glyph_extents (PangoFont        *font,
     }
   else
     {
+    fallback:
+      
       if (ink_rect)
 	{
 	  ink_rect->x = 0;
@@ -702,10 +688,8 @@ load_fallback_font (PangoXftFont *xfont)
   _pango_xft_font_map_get_info (xfont->fontmap, &display, &screen);
       
   xft_font = XftFontOpen (display,  screen,
-                          XFT_FAMILY, XftTypeString, "sans",
-                          XFT_ENCODING, XftTypeString, "glyphs-fontspecific",
-                          XFT_CORE, XftTypeBool, False,
-                          XFT_SIZE, XftTypeDouble, (double)pango_font_description_get_size (xfont->description)/PANGO_SCALE,
+                          FC_FAMILY, FcTypeString, "sans",
+                          FC_SIZE, FcTypeDouble, (double)pango_font_description_get_size (xfont->description)/PANGO_SCALE,
                           NULL);
   
   if (!xft_font)
@@ -714,15 +698,36 @@ load_fallback_font (PangoXftFont *xfont)
       exit (1);
     }
 
-  if (!set_unicode_charmap (xft_font->u.ft.font->face))
-    {
-      g_warning ("Cannot set unicode character map for fallback font, nothing to do");
-      exit (1);
-    }
-
   xfont->xft_font = xft_font;
 }
 
+static XftFont *
+xft_font_get_font (PangoFont *font)
+{
+  PangoXftFont *xfont;
+  Display *display;
+  int screen;
+
+  xfont = (PangoXftFont *)font;
+
+  if (xfont->xft_font == NULL)
+    {
+      _pango_xft_font_map_get_info (xfont->fontmap, &display, &screen);
+
+      xfont->xft_font = XftFontOpenPattern (display, FcPatternDuplicate (xfont->font_pattern));
+      if (!xfont->xft_font)
+	{
+	  gchar *name = pango_font_description_to_string (xfont->description);
+	  g_warning ("Cannot open font file for font %s", name);
+  	  g_free (name);
+  	  
+ 	  load_fallback_font (xfont);
+	}
+    }
+  
+  return xfont->xft_font;
+}
+
 /**
  * pango_xft_font_get_font:
  * @font: a #PangoFont.
@@ -734,51 +739,9 @@ load_fallback_font (PangoXftFont *xfont)
 XftFont *
 pango_xft_font_get_font (PangoFont *font)
 {
-  PangoXftFont *xfont;
-  Display *display;
-  int screen;
-
   g_return_val_if_fail (PANGO_XFT_IS_FONT (font), NULL);
   
-  xfont = PANGO_XFT_FONT (font);
-
-  if (xfont->xft_font == NULL)
-    {
-      _pango_xft_font_map_get_info (xfont->fontmap, &display, &screen);
-
-      xfont->xft_font = XftFontOpenPattern (display, xfont->font_pattern);
-      if (!xfont->xft_font)
-	{
-	  gchar *name = pango_font_description_to_string (xfont->description);
-	  
-	  /* No provision for failure here, unfortunately, various fonts
-	   * fail only when FT tries to open them... e.g., fonts with only
-	   * bitmaps.
-	   */
-	  g_warning ("Cannot open font file for font %s", name);
-	  g_free (name);
-	  
-	  load_fallback_font (xfont);
-	}
-      else
-	{
-	  /* There should be a unicode encoding, since we queried for it */
-	  if (!set_unicode_charmap (xfont->xft_font->u.ft.font->face))
-	    {
-	      gchar *name = pango_font_description_to_string (xfont->description);
-	      
-	      g_warning ("Cannot load unicode character map for font %s", name);
-	      g_free (name);
-	      
-	      XftFontClose (display, xfont->xft_font);
-	      xfont->xft_font = NULL;
-		
-	      load_fallback_font (xfont);
-	    }
-	}
-    }
-
-  return xfont->xft_font;
+  return xft_font_get_font (font);
 }
 
 /**
@@ -823,54 +786,89 @@ pango_xft_font_get_unknown_glyph (PangoFont *font,
 }
 
 /**
- * pango_xft_font_get_face:
+ * pango_xft_font_lock_face:
  * @font: a #PangoFont.
  *
- * Gets the FreeType FT_Face associated with a font.
+ * Gets the FreeType FT_Face associated with a font,
+ * This face will be kept around until you call
+ * pango_xft_font_unlock_face().
  *
  * Returns: the FreeType FT_Face associated with @font.
  **/
 FT_Face
-pango_xft_font_get_face (PangoFont *font)
+pango_xft_font_lock_face (PangoFont *font)
 {
   XftFont *xft_font;
 
   g_return_val_if_fail (PANGO_XFT_IS_FONT (font), NULL);
 
-  xft_font = pango_xft_font_get_font (font);
+  xft_font = xft_font_get_font (font);
   
-  if (xft_font->core)
-    return NULL;
-  else
-    return xft_font->u.ft.font->face;
+  return XftLockFace (xft_font);
 }
 
 /**
- * pango_xft_font_get_ot_info:
+ * pango_xft_font_unlock_face:
  * @font: a #PangoFont.
  *
- * Gets the OpenType info of a font as a #PangoOTInfo.
- * 
- * Returns: the OpenType info of @font, or %NULL if there is none.
+ * Releases a font previously obtained with
+ * pango_xft_font_lock_face().
  **/
-PangoOTInfo *
-pango_xft_font_get_ot_info (PangoFont *font)
+void
+pango_xft_font_unlock_face (PangoFont *font)
 {
-  PangoXftFont *xfont;
+  XftFont *xft_font;
 
-  g_return_val_if_fail (PANGO_XFT_IS_FONT (font), NULL);
+  g_return_if_fail (PANGO_XFT_IS_FONT (font));
 
-  xfont = PANGO_XFT_FONT (font);
+  xft_font = xft_font_get_font (font);
+  
+  XftUnlockFace (xft_font);
+}
 
-  if (!xfont->ot_info)
-    {
-      FT_Face face = pango_xft_font_get_face (font);
-      
-      if (!face)
-	return NULL;
+/**
+ * pango_xft_font_get_glyph:
+ * @font: a #PangoFont for the Xft backend
+ * @wc: Unicode codepoint to look up
+ * 
+ * Gets the glyph index for a given unicode codepoint
+ * for @font. If you only want to determine
+ * whether the font has the glyph, use pango_xft_font_has_char().
+ * 
+ * Return value: the glyph index, or 0, if the unicode
+ *  codepoint doesn't exist in the font.
+ **/
+guint
+pango_xft_font_get_glyph (PangoFont *font,
+			  gunichar   wc)
+{
+  XftFont *xft_font;
 
-      xfont->ot_info = pango_ot_info_new (face);
-    }
+  g_return_val_if_fail (PANGO_XFT_IS_FONT (font), 0);
 
-  return xfont->ot_info;
+  xft_font = xft_font_get_font (font);
+  
+  return XftCharIndex (NULL, xft_font, wc);
+}
+
+/**
+ * pango_xft_font_has_char:
+ * @font: a #PangoFont for the Xft backend
+ * @wc: Unicode codepoint to look up
+ * 
+ * Determines whether @font has a glyph for the codepoint @wc.
+ * 
+ * Return value: %TRUE if @font has the requested codepoint.
+ **/
+gboolean
+pango_xft_font_has_char (PangoFont *font,
+			 gunichar   wc)
+{
+  XftFont *xft_font;
+
+  g_return_val_if_fail (PANGO_XFT_IS_FONT (font), 0);
+
+  xft_font = xft_font_get_font (font);
+  
+  return XftCharExists (NULL, xft_font, wc);
 }
